/* xigc.c
 * S.Coffin	GV Computing	1/93
 *
 * Main X11 program for displaying Goban widget, menus, and etc
 */

/*  Copyright (c) 1992,1993,1994,1995 by Stephen Coffin.  All rights reserved.
 *
 *  This program is distributed in the hope that it will be useful.
 *  Use and copying of this software and preparation of derivative works
 *  based upon this software are permitted, so long as the following
 *  conditions are met:
 *       o credit to the authors is acknowledged following current
 *         academic behavior
 *       o no fees or compensation are charged for use, copies, or
 *         access to this software
 *	 o no fees or compensation are charged for access to the
 *	   game server for which this software is a client
 *       o this copyright notice is included intact.
 *  This software is made available AS IS, and no warranty is made about 
 *  the software or its performance. 
 * 
 *  Bug descriptions, use reports, comments or suggestions are welcome.
 *  Send them to    scoffin@netcom.com
 */

/* uncomment this if you want button 2 to backup a move instead of lifting 
 * a stone off the game board.
 */
/*JL #define NOLIFTSTONE JL*/

#include <stdio.h>

#ifdef DEF_X11R3
#include <Intrinsic.h>
#include <Shell.h>
#include <StringDefs.h>
#include <Xlib.h>
#include <Xutil.h>
#include <cursorfont.h>
#include <Command.h>
#include <Form.h>
#include <Box.h>
#include <Dialog.h>
#include <AsciiText.h>
#define XP char *
#else
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/AsciiText.h>
#define XP XtPointer
#endif
#include "filter.h"
#include "shared.h"
#include "stipple.bm"
#include "xigc_icon.bm"

/* the default communications command to execute */
/*#define  CMD "igscomm -h 128.218.30.183 -p 6969"      /* BERKELEY */
#define  CMD "igscomm -h 203.255.112.3 -p 6969"	/* igs.nuri.net */
/* the command to kill the communications proc */
#define QUITCMD "."

/* Use of telnet not recommended =SC */
/* #define CMD "telnet 203.255.112.3 6969" */
/* #define QUITCMD "\nquit" */

/* #define CMD "rlogin uswat" */
/* #define QUITCMD "~." */

/* #define CMD "cu uswat" */
/* #define QUITCMD "~." */

#define USAGE "%s [x-arg-list] [-?] [-c \"command line to execute\"] [-d] [-s] [-a] [-l] [-f filename]\n"
#define TITLE "xigc Menu"
#define ICONTITLE "xigc Menu"
#define BOARDTITLE "xigc"
#define BOARDICON "xigc Board"

unsigned char cmdline[80], fileline[80];
extern int move;
extern int stepmove;
int currentmove;
unsigned char stippleflag = FALSE;
unsigned char cmd1[80], cmd2[80], quitcmd[80];
static void ClearBoard();
int tersemode = FALSE;
int sortwhomode = FALSE;
extern unsigned char my_moveflag;
extern int putout();
extern int filterstatus;
extern void HandleStringEvent(), StringInput();
extern char prompt_str[];
extern long ots, nts;

#define WATCHDOGINTERVAL 400
static int watchdog = 0;
int hiliteb=FALSE, hilitew=FALSE;
int hilitel=FALSE, hiliteh=FALSE;

#include "Goban.h"
void SetCursor ();
void read_filter(), read_kbd(), read_pipe();
int setup_filter();
int input0id, input1id, input2id;
int i0id, i1id, i2id;
XtAppContext app_context;
int dbg = FALSE;
static int cmdflag = FALSE;
static int saveformat = SGF;
static int get_cmd();
static int update_int = 1;
static int timer_beep = 0;
static int timer_color = 30;
int boardchanged = FALSE;
int mouseflag = FALSE;
unsigned char sgflocalflag = FALSE, localflag = FALSE;
unsigned char teachflag = FALSE;
char sgflocalfile[LINESIZE];
FILE sgf_file_fp;
static unsigned char liftflag = FALSE;
unsigned char statusflag = FALSE;
unsigned char savedir[80];
int clickflag = 0; /* If one use audio driver for click */

#include "goprocs.h"

game maingame;
game stepgame;

/* these fp's are the std i/o when we EXEC a process */
FILE *exec_stdin=NULL, *exec_stdout=NULL, *exec_stderr=NULL;
int gameno=(-1);
char ffre=' ', rules= 'I';
int handicap = 0;
static int fast_step_size = 20;
extern ObsState observeflag;
int boardsize = 19;
unsigned char player1[40], player2[40];
int p1_sec, p2_sec;
int p1_captured, p2_captured;
int p1_byoyomi, p2_byoyomi;
extern char rank1[], rank2[];
extern float komi;

Widget toplevel, bw, goban, refresh, clearb, form, size, coordinates;
Widget stepf, stepff, stepb, stepbf, refreshstep, savecurrent;
Widget cursorcolor, savegame, localrefresh, quit, echow;
Widget gamew, hcapw, whitew, blackw, captw, captb, timew, timeb, lastw;
Widget sgfloadw, sgfdonew, dialog, dp, textwidget;
Widget whow, gamesw, passw, undow;

#define MAXSIZE 19
int colorflag = BLACK;

beep() {
    XBell( XtDisplay(toplevel), 0 );
}

highlight( whichone, flag )
	Widget whichone;
	int flag;
{
	int j, doit;
	Arg args[2];
	Pixel fg, bg, xx;
	static Pixel bg_default = (-1);

	if( teachflag && flag == TRUE ) return;

	/* get the default background color if we do not know it */
	if( bg_default == (-1) ) {
		j = 0;
		XtSetArg( args[j], XtNbackground, &bg_default ); ++j;
		XtGetValues( toplevel, args, j );
	}

	/* get current hilite state for this widget */
	j = 0;
	XtSetArg( args[j], XtNforeground, &fg ); ++j;
	XtSetArg( args[j], XtNbackground, &bg ); ++j;
	XtGetValues( whichone, args, j );

	/* now unhighlited */
	if( bg == bg_default ) {
		if( flag ) doit = TRUE;
		else doit = FALSE;
	}

	/* now highlited */
	else {
		if( flag ) doit = FALSE;
		else doit = TRUE;
	}

	/* switch `em */
	if( doit ) {
		xx = fg;
		fg = bg;
		bg = xx;
		j = 0;
		XtSetArg( args[j], XtNforeground, fg ); ++j;
		XtSetArg( args[j], XtNbackground, bg ); ++j;
		XtSetValues( whichone, args, j );
	}
}

void local_highlight( flag )
	int flag;
{
	highlight( localrefresh, flag );
	hilitel = flag;
}

void host_highlight( flag )
	int flag;
{
	highlight( refresh, flag );
	hiliteh = flag;
}

void time_highlight( flag )
	int flag;
{
	highlight( timew, flag );
	highlight( timeb, flag );
	hilitew = flag;
	hiliteb = flag;
}

clearit() {
	GbClearBoard( goban );
}

void SgfLoad( w, client_data, call_data )
	Widget w;
	XP client_data;
	XP call_data;
{
	Arg arg[10];
	int j;
	Position x, y, xx, yy;

	if( dbg ) fprintf( stderr, "SgfLoad widget entered\n" );
	if( observeflag == PLAYING || observeflag == SCORING || teachflag) {
	    beep();
	    fprintf( stderr, "LOCAL: cannot load file while playing or reviewing\n" );
	    fprintf( exec_stdin, "\n" );
	    fflush( exec_stdin );
	    return;
	}

        /* translate coords in application top-level window */
    	/* into coords from root window origin.	*/
    	XtTranslateCoords( toplevel, (Position)20, (Position)20, &xx, &yy );

	j = 0;
	XtSetArg( arg[j], XtNheight, (Dimension)80 ); ++j;
	XtSetArg( arg[j], XtNlabel, "SGF File to load:" ); ++j;
	XtSetArg( arg[j], XtNvalue, "" ); ++j;
	XtSetValues( dialog, arg, j );

	XtPopup( dp, XtGrabNonexclusive );

	textwidget = XtNameToWidget( dialog, "value");
	XtOverrideTranslations( textwidget, XtParseTranslationTable(
					"<Key>Return: nline()") );

	j = 0;
	XtSetArg( arg[j], XtNx, xx ); ++j;
	XtSetArg( arg[j], XtNy, yy ); ++j;
	XtSetValues( dp, arg, j );
}

void Cancel( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	if( dbg ) fprintf( stderr, "Cancel callback entered\n" );
	XtPopdown( dp );
}

void Confirm( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	String s;
	char ss[LINESIZE];

	if( dbg ) fprintf( stderr, "Confirm callback entered\n" );

	s = XawDialogGetValueString( dialog );
	XtPopdown( dp );

	if( dbg ) fprintf( stderr, "\tString >>%s<<\n", s );

	if( strlen( s ) == 0 )
	    fprintf( stderr, "LOCAL: Must give filename of SGF file\n" );
	else {
	    sprintf( ss, "sgfload %s\n", s );
	    process_kbd( ss, exec_stdin );
		if( localflag ) {
			fprintf( stderr, "%s> ", prompt_str );
			fflush( stderr );
		}
	}
}

void make_undo_sensitive( which )
	Boolean which;
{
	int j;
	Arg args[2];

    XtSetArg( args[0], XtNsensitive, which );
    XtSetValues( passw, args, 1 );
    XtSetValues( undow, args, 1 );
}

void SgfDone( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	if( dbg ) fprintf( stderr, "SgfDone widget entered\n" );
	process_kbd( "sgfdone\n", exec_stdin );
	if( localflag ) {
		fprintf( stderr, "%s> ", prompt_str );
		fflush( stderr );
	}
}

void Quit( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	if( localflag ) exit( 0 );
	if( cmdflag == TRUE ) {
	    beep();
	    fprintf( stderr,
		"\rLOCAL: you must log out of IGS and each intermedate system manually\n" );
	    fprintf( stderr, " when using the \"-c\" option\n" );
	    fprintf( exec_stdin, "\n" );
	    fflush( exec_stdin );
	    return;
	}

	fprintf( exec_stdin, "quit\n" );
	fflush( exec_stdin );
	sleep( 1 );

	if( quitcmd[0] == '\0' )
		fprintf( exec_stdin, "\n%s\n", QUITCMD );
	else
		fprintf( exec_stdin, "\n%s\n", quitcmd );
	fflush( exec_stdin );
	sleep( 1 );
	rdwr_pclose( NULL, NULL, NULL );
	exit( 0 );
}

static void SaveGame( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	unsigned char s[40], s2[40];
	int ret;
	struct stat sbuf;

	if( dbg )
	    fprintf( stderr, "IN SAVEGAME call_data=0x%x\n", client_data );

	if(client_data == (XP)REALMOVE && 
	     (gameno < 0 || maingame.stoneNumber < 1) ) {
	    beep();
	    fprintf( stderr, "\rLOCAL:  No active game to save\n" );
	    fflush( stderr );
	}
	else if(client_data == (XP)STEPMOVE && 
	     stepgame.stoneNumber < 1 ) {
	    beep();
	    fprintf( stderr, "\rLOCAL:  No active game to save\n" );
	    fflush( stderr );
	}
	else if( client_data == (XP)STEPMOVE &&
	     stepgame.stoneNumber >= Maxstones) {
	    beep();
	    fprintf( stderr, "\rLOCAL:  No changes to current game:  use \"Save Game\" instead\n" );
	    fflush( stderr );
	}
	else if( client_data == (XP)STEPMOVE && liftflag ) {
	    beep();
	    fprintf( stderr, "\rLOCAL:  Cannot \"Save to Current\" after a stone has been lifted with button 2\n" );
	    fflush( stderr );
	}
	else {
	    if( (int)strlen( (char *)player1 ) <= 0 || (int)strlen( (char *)player2 ) <= 0 )
		strcpy( (char *)s, "no_players" );
	    else sprintf( (char *)s, "%s-%s", player1, player2 );

	    /* find an unused filename */
	    sprintf( (char *)s2, "%s/%s", savedir, s );
	    for( ret = 1 ;; ++ret ) {
		if( stat( (char *)s2, &sbuf ) == 0 ) {
		    sprintf( (char *)s2, "%s/%s.%d", savedir, s, ret );
		    continue;
		}
		else break;
	    }

	    if( saveformat == SGF )
	        ret = save_game_sgf( s2, player1, player2, handicap, rank1,
					rank2, komi, client_data );
	    else
	        ret = save_game( s2, player1, player2, handicap, client_data );
	    if( ret ) {
	       fprintf( stderr, "\rLOCAL:  Game saved to %s\n", s2 );
	       fflush( stderr );
	    }
	}
	fprintf( exec_stdin, "\n" );
	fflush( exec_stdin );
	return;
}

set_cursor_color( m )
	int m;
{
	if( m % 2 ) {
	    colorflag = BLACK;
	    SetCursor( GbCBlackStone );
	}
	else {
	    colorflag = WHITE;
	    SetCursor( GbCWhiteStone );
	}
	return( colorflag );
}
		

void do_refresh( m, type )
	int m;
	int type;
{
	int i, x, y, flag;

	GbClearBoard( goban );
	update_handicap( handicap, komi );

	if (maingame.stoneNumber) {
	  drawall(&maingame);
	}
	stepgame.stoneNumber = Maxhistory;

	set_cursor_color( m );
}

void RefreshStep( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	if (stepgame.stoneNumber < Maxhistory) {
	  GbClearBoard( goban );

/* Back off all stepgame stones until the real game is found - that's the 
 * "step" point. 
 */
	  while ((stepgame.stoneNumber > 0) && 
		 (stepgame.sgf.gametype[stepgame.stoneNumber] != REALGAME)) {
	    backup(&stepgame,stepgame.stoneNumber,DRAWIT,((rules==IRULES)?RULES:GOE_RULES));
	    backupend(&stepgame,stepgame.stoneNumber - 1,DRAWIT,0);
	  }
	  drawall(&stepgame);

/* If the stepgame is the same as the maingame, then switch to the main */
	  if ((stepgame.sgf.gametype[stepgame.stoneNumber]==REALGAME) &&
	      (maingame.stoneNumber == stepgame.stoneNumber)) {
	    stepgame.stoneNumber = Maxhistory;
	    boardchanged = FALSE;
	    mouseflag = FALSE;
	    liftflag = FALSE;
	    highlight( localrefresh, FALSE );
	  } else {
	    boardchanged = TRUE;
	    mouseflag = FALSE;
	  }
	} else
	  beep();
}

void LocalRefresh( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	int i;
	int x, y;

	liftflag = boardchanged = mouseflag = statusflag = FALSE;

	do_refresh( move, REALMOVE );
	stepmove = currentmove = move;
	highlight( localrefresh, FALSE );
	update_status( gameno, rules, ffre, player2, player1, p2_captured,
		p1_captured,
		p2_sec, p1_sec, p2_byoyomi, p1_byoyomi, rank1, rank2 );
	update_handicap( handicap, komi );
}

unsigned char keyecho = TRUE;

/* handle keystroke events */
static void keyevent( w, event )
	Widget w;
	XEvent *event;
{
	int i, j, len;
	XComposeStatus compose;
	KeySym mykey;
	static unsigned char keytext[32];
	char *txt;
	static char sendtext[256];
	Arg args[2];

	i = XLookupString( &event->xkey, (char *)keytext, 32, &mykey, &compose );

	if( dbg ) fprintf( stderr,
		"In keyevent KEYPRESS [i=%d, mykey=0x%x]  keytext=>>%s<<\n",
		i, mykey, keytext );

	/* newline: send line to host */
	if( keytext[0] == '\n' || keytext[0] == '\r') {

		watchdog = 0;

        XtSetArg( args[0], XtNstring, &txt );
        XtGetValues( echow, args, 2 );

        /* got a line, now process it and send */

        /* first, replace any newlines in string put in spuriously
         * by text widget with spaces
         */
        len = strlen( txt );
        for( i = 0; i < len; ++i ) {
            if( txt[i] != '\n' ) sendtext[i] = txt[i];
            else sendtext[i] = ' ';
		}

        sendtext[i++] = '\n';
        sendtext[i++] = '\0';
        process_kbd( sendtext, exec_stdin );
        if( keyecho ) fprintf( stderr, "%s", sendtext );

        j = 0;
        sendtext[0] = '\0';
        XtSetArg( args[j], XtNstring, sendtext ); ++j;
		if( !keyecho ) XtSetArg( args[j], XtNecho, True ); ++j;
        XtSetValues( echow, args, j );
        keyecho = TRUE;
	}

	/* CNTRL-P: turn off echo */
	else if( keytext[0] == 0x10 ) {
	    if( keyecho ) {
	        keyecho = FALSE;
        	XtSetArg( args[0], XtNecho, False );
            XtSetValues( echow, args, 1 );
	    }
	    else {
	        keyecho = TRUE;
        	XtSetArg( args[0], XtNecho, True );
            XtSetValues( echow, args, 1 );
	    }
	}

	/* all other keystrokes: forward to text widget */
	else {
		((XKeyEvent *)event)->window = XtWindow( echow );
		XtDispatchEvent( event );
        keyecho = TRUE;
	}

	fflush( stderr );
}

static void CursorColor( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	boardchanged = mouseflag = TRUE;
	highlight( localrefresh, TRUE );
	set_cursor_color( colorflag+1 );
}

static void StepForward( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	int i, x, y, flag, size;
  	static char fwd[] = "forward\n";

	if (observeflag == SCORING) {
	  beep();
	  return;
	}

	if( dbg )
	    fprintf( stderr, "StepForward entered [stepmove=%d, move=%d]\n",
			stepmove, move );

	if( client_data ) size = fast_step_size;
	else size = 1;

	for( ; size > 0; --size ) {

	/* Is there another move that can be made? */
	if( humanforward( &stepgame ) ) {

/* If the stepgame is the same as the maingame, then switch to the main */
	  if ((stepgame.sgf.gametype[stepgame.stoneNumber] == REALGAME) &&
	      (maingame.stoneNumber == stepgame.stoneNumber)) {
	    stepgame.stoneNumber = Maxhistory;
	    boardchanged = liftflag = mouseflag = FALSE;
	    highlight( localrefresh, FALSE );
	  }
	} else if( teachflag ) {
	  process_kbd( fwd, exec_stdin );
	if( localflag ) {
		fprintf( stderr, "%s> ", prompt_str );
		fflush( stderr );
	}
	} else if( sgflocalflag ) {
	    process_local_sgf_line( 1 );
	} else {
	  if ((stepgame.sgf.gametype[stepgame.stoneNumber] == REALGAME) &&
	      (maingame.stoneNumber > stepgame.stoneNumber) &&
	      (stepgame.stoneNumber < (Maxhistory-1))) {
	    stepgame.who = maingame.sgf.typehistory[stepgame.stoneNumber+1] & 1;
	    addastone(&stepgame,
		      maingame.history[stepgame.stoneNumber+1],
		      DRAWIT, TRUE, Playstone, ((rules==IRULES)?RULES:GOE_RULES));

	    if (maingame.stoneNumber == stepgame.stoneNumber) {
	      stepgame.stoneNumber = Maxhistory;
	      mouseflag = boardchanged = liftflag = FALSE;
	      setcursorcolor(&maingame);
	      highlight( localrefresh, FALSE );
	    } else {
	      setcursorcolor(&stepgame);
	      highlight( localrefresh, TRUE );
	    }

	  } else if (stepgame.sgf.gametype[stepgame.stoneNumber] != REALGAME) {
	    beep();
	    break;
	  } else {
	    stepgame.stoneNumber = Maxhistory;
	    boardchanged = mouseflag = liftflag = FALSE;
	    highlight( localrefresh, FALSE );
	    beep();
	    break;
	} } }
}

static void StepBack( w, client_data, call_data )
	GobanWidget w;
	XP client_data;
	XP call_data;
{
	int size;
	char bkw[20];

	if( client_data ) size = fast_step_size;
	else size = 1;

	if( teachflag ) {
	  	if( client_data ) sprintf( bkw, "undo %d\n", fast_step_size );
	  	else sprintf( bkw, "undo\n" );
	  	process_kbd( bkw, exec_stdin );
	  	return;
	}

	for( ; size > 0; --size ) {

	if (maingame.stoneNumber && (stepgame.stoneNumber >= Maxhistory)) {
	  stepgame = maingame;
	  stepgame.gametype = STEPGAME;
	}

	if (stepgame.stoneNumber && (stepgame.stoneNumber < Maxhistory)) {

/* Maybe you would like to backup a stone without uncapturing stones that 
 * were captured by the original play.  In that case use STEP_RULES here 
 * instead of RULES, but code must also be added to change the "gametype" 
 * of every move in the stepgame to STEPGAME - otherwise it is possible for 
 * the stepgame to be at odds with maingame without the programming knowing. 
 * This change would mean that the user would be required to select "Refresh 
 * Local" anytime the user backs up a stone that caused captures - 
 * which of course redraws all of the stones on the game board.  Another 
 * possibility is to add a function that compares the stepgame to the
 * maingame so that the program knows when they are in sync.
 */
	  backup(&stepgame,stepgame.stoneNumber,DRAWIT,((rules==IRULES)?RULES:GOE_RULES));
	  backupend(&stepgame,stepgame.stoneNumber - 1,DRAWIT,0);

/* If the stepgame is the same as the maingame, then switch to the main */
	  if ((stepgame.sgf.gametype[stepgame.stoneNumber]==REALGAME) &&
	      (maingame.stoneNumber == stepgame.stoneNumber)) {
	    stepgame.stoneNumber = Maxhistory;
	    boardchanged = mouseflag = liftflag = FALSE;
	    highlight( localrefresh, FALSE );
	  } else {
	    boardchanged = TRUE;
	    highlight( localrefresh, TRUE );
	  }

	} else {
	  beep();
	}
	}
}

static void InputStone( w, event )
	GobanWidget w;
	XButtonEvent *event;
{
	Position x = event->x;
	Position y = event->y;

	if( GbGetStoneFromPosition( w, &x, &y ) == TRUE ) {
	    if( dbg ) fprintf( stderr, "Input stone at %c%d\n",
				'A' + (x > 8 ? x + 1 : x) - 1, y );
	}
	else {
	    beep();
	    return;
	}

	if( observeflag != PLAYING && observeflag != SCORING ) {
	    beep();
	    return;
	}
	else if( observeflag != SCORING && (boardchanged || mouseflag) ) {
	    beep();
	    return;
	}

	nts = time( NULL );
	fprintf( exec_stdin, "%c%d %d\n", 'A' + (x > 8 ? x + 1 : x) - 1, y,
			(int)( nts - ots ) );
	fflush( exec_stdin );
	if( !tersemode ) {
	    putout( "\n", FALSE, FALSE, 46 );
	}
	my_moveflag = TRUE;
}

static void AlternateStone( w, event )
    GobanWidget w;
    XButtonEvent *event;
{
	int ob;
	Position x = event->x;
	Position y = event->y;

	if( GbGetStoneFromPosition( w, &x, &y ) == TRUE ) {
	    if( dbg ) fprintf( stderr, "Alternate stone at %c%d\n",
				'A' + (x > 8 ? x + 1 : x) - 1, y );
	}
	else {
	    beep();
	    return;
	}

	if( statusflag || (observeflag == SCORING)) {
	    beep();
	    return;
	}

	++currentmove;
	if( gameno < 0 ) gameno = 0;
	boardchanged = mouseflag = TRUE;

/* Check to see if the stepgame is already in use.  If not, then either
 * add to the maingame (if there is one) or prepare to use the stepgame
 */
	if (stepgame.stoneNumber >= Maxhistory) {       /* Stepgame in use? */
	    stepgame = maingame;                        /* Use maingame. */
	    stepgame.gametype = STEPGAME;
	}

	if( colorflag == BLACK ) {
	    makemove(x,y,currentmove,(int)GbBlackStone,STEPMOVE,DRAWIT,((rules==IRULES)?STEP_RULES:GOE_STEP_RULES));
	}
	else {
	    makemove(x,y,currentmove,(int)GbWhiteStone,STEPMOVE,DRAWIT,((rules==IRULES)?STEP_RULES:GOE_STEP_RULES));
	}

	nomorehistory(&stepgame);
	highlight( localrefresh, TRUE );
}


static void RemoveStone( w, event )
    GobanWidget w;
    XButtonEvent *event;
{
	int s1;
	Position x;
	Position y;

	if (observeflag == SCORING) {
	  beep();
	  return;
	}

#ifdef NOLIFTSTONE
	if (maingame.stoneNumber && (stepgame.stoneNumber >= Maxhistory)) {
	  stepgame = maingame;
	  stepgame.gametype = STEPGAME;
	}

	if (stepgame.stoneNumber && (stepgame.stoneNumber < Maxhistory)) {
	  backup(&stepgame,stepgame.stoneNumber,DRAWIT,((rules==IRULES)?RULES:GOE_RULES));
	  backupend(&stepgame,stepgame.stoneNumber - 1,DRAWIT,0);

/* If the stepgame is the same as the maingame, then switch to the main */
	  if ((stepgame.sgf.gametype[stepgame.stoneNumber]==REALGAME) &&
	      (maingame.stoneNumber == stepgame.stoneNumber)) {
	    stepgame.stoneNumber = Maxhistory;
	    boardchanged = FALSE;
	    mouseflag = FALSE;
	    highlight( localrefresh, FALSE );
	  } else
	    boardchanged = TRUE;

	} else
	  beep();
#else
	x = event->x;
	y = event->y;

	if (GbGetStoneFromPosition (w, &x, &y) == TRUE) {
	    if( dbg ) fprintf ( stderr, "Remove stone at %c%d\n",
					'A' + (x > 8 ? x + 1 : x) - 1, y);
	}

	if( statusflag ) {
	    beep();
	    return;
	}

	s1 = y-1+(maingame.width*(x-1));

	if (maingame.stoneNumber && (stepgame.stoneNumber >= Maxhistory)) {
	  stepgame = maingame;
	  stepgame.gametype = STEPGAME;
	}

	if (stepgame.board[s1] != Blank) {
	  liftastone(&stepgame,s1,STEPGAME);
	  boardchanged = TRUE;
	  liftflag = TRUE;
	  highlight( localrefresh, TRUE );
	} else
	  beep();
#endif
}

void Ticker( client_data, timer )
	XP client_data;
	XtIntervalId *timer;
{
	XtAppAddTimeOut( (XtAppContext)client_data, 1000, Ticker, client_data );

#ifdef SC
	if( ++watchdog > WATCHDOGINTERVAL && observeflag != PLAYING ) {
	    fprintf( exec_stdin, "ayt\n" );
	    fflush( exec_stdin );
	    watchdog = 0;
	}
#endif

	if( observeflag == SCORING || observeflag == NOOBSERVE ||
	    move < 0 || statusflag ) return;

	else if( move % 2 ) {
	    if( FALSE ) fprintf( stderr, "TIMER EVEN [B]\n" );
	    if( !(--p2_sec % update_int)) {
		fix_btime( p2_sec, p2_byoyomi );
	        if( observeflag == PLAYING &&
		    timer_beep != 0 &&
		    (p2_sec >= 0 ) &&
		    (p2_sec <= timer_beep) ) beep();
	    }
	}
	else {
	    if( FALSE ) fprintf( stderr, "TIMER ODD [W]\n" );
	    if (!(--p1_sec % update_int)) {
		fix_wtime( p1_sec, p1_byoyomi );
	        if( observeflag == PLAYING &&
		    timer_beep != 0 &&
		    (p1_sec >= 0 ) &&
		    (p1_sec <= timer_beep) ) beep();
	   }
	}
}

void SetCursor ( cs )
	int cs;
{
	Arg args[1];

	switch (cs) {
	    case GbCBlackStone : 
		XtSetArg   (args[0], XtNcursor, GbCBlackStone);
		if( dbg ) fprintf ( stderr, "Set cursor to : Black \n"); 
		break;
	    case GbCWhiteStone : 
		XtSetArg   (args[0], XtNcursor, GbCWhiteStone );
		if( dbg ) fprintf ( stderr, "Set cursor to : White\n"); 
		break;
	    case XC_question_arrow :
		XtSetArg   (args[0], XtNcursor, XC_question_arrow );  
		if( dbg ) fprintf ( stderr, "Set cursor to : Question\n"); 
		break;
	}

	XtSetValues (goban, args, 1);
}

void SendMsg( w, client_data, call_data )
    Widget w;
    XP client_data;
    XP call_data;
{
	if( dbg ) fprintf(stderr, "Sending:  %s", client_data );
	process_kbd( client_data, exec_stdin );
}

void RefreshHost( w, client_data, call_data )
    GobanWidget w;
    XP client_data;
    XP call_data;
{
	if( gameno < 0 ) {
	    beep();
	    highlight( localrefresh, FALSE );
	    fprintf( stderr, "\rLOCAL:  No active game to refresh\n" );
	    fflush( stderr );
	    highlight( localrefresh, FALSE );
	    GbClearBoard( goban );
	    initboard( boardsize );
	    fprintf( exec_stdin, "\n" );
	    fflush( exec_stdin );
	}
	else {
	    SetCursor ( GbCBlackStone );
	    move = (-1);
	    refreshboard();
	    fprintf( exec_stdin, "moves %d\n", gameno );
	    fflush( exec_stdin );
	}

	highlight( refresh, FALSE );
}

int refreshboard() {
	if( gameno < 0 ) {
	    highlight( localrefresh, FALSE );
	    fprintf( stderr, "\rLOCAL: No active game to refresh\n" );
	    fflush( stderr );
	    fprintf( exec_stdin, "\n" );
	    fflush( exec_stdin );
	    return( 0 );
	}

	boardchanged = mouseflag = statusflag = liftflag = FALSE;
	highlight( localrefresh, FALSE );
	GbClearBoard( goban );
	stepmove = (-1);
	currentmove = (-1);
	handicap = 0;
	initboard( boardsize );
	return( 0 );
}

static void ClearBoard() {
short i;
	GbClearBoard( goban );
	if (maingame.stoneNumber && (stepgame.stoneNumber >= Maxhistory)) {
	  stepgame = maingame;
	  stepgame.gametype = STEPGAME;
	}
	stepgame.stoneNumber = 0;	/* Beginning of game */

	/* Clear the board, but leave the history so we can step through */
	for (i = 0; i < Maxstones; i++) {
	  stepgame.board[i] = Blank;
	  stepgame.brdtohst[i] = 0;
	}
	allarmies(&stepgame);

	boardchanged = TRUE;
	liftflag = FALSE;
	mouseflag = FALSE;
	highlight( localrefresh, TRUE );
	set_cursor_color( 1 );
}

void SetSize( w, client_data, call_data )
    GobanWidget w;
    XP client_data;
    XP call_data;
{
	Arg args[1];
	static int beenhere = FALSE;

	if( !beenhere ) boardsize = 13;

	if( call_data == NULL ) {
	    switch( boardsize ) {
	    case 9:
	        boardsize = 13;
	        break;
	    case 13:
	        boardsize = 19;
	        break;
	    case 19:
	    default:
	        boardsize = 9;
	        break;
	    }
	}

	if( dbg ) fprintf ( stderr, "Set size to : %d\n", boardsize );

	if( beenhere ) {
	    if( observeflag && call_data == NULL ) {
		boardchanged = mouseflag = TRUE;
		highlight( localrefresh, TRUE );
	    }
	    else {
		boardchanged = mouseflag = liftflag = FALSE;
	        highlight( localrefresh, FALSE );
	    }
            stepmove = (-2);
	    currentmove = (-1);
	}
	else beenhere = TRUE;

	set_cursor_color( 1 );

	XtSetArg( args[0], XtNgameSize, (Dimension)boardsize );
	XtSetValues( goban, args, 1 );
}

static void SetCoordinates() {
	Arg     args[1];
	Boolean coordinates;

	XtSetArg    (args[0], XtNdisplayCoordinates, &coordinates);
	XtGetValues (goban, args, 1);
	
	switch (coordinates) {
	    case FALSE : 
		XtSetArg (args[0], XtNdisplayCoordinates, TRUE) ; 
		if( dbg ) fprintf( stderr, "Set coordinates to : TRUE \n");
		break;
	    case TRUE  : 
		XtSetArg (args[0], XtNdisplayCoordinates, FALSE); 
		if( dbg ) fprintf( stderr, "Set coordinates to : FALSE \n");
		break;
	}

	XtSetValues (goban, args, 1);
}

static void SetupBoard() {
	unsigned char s[40];
	
	XtPopup( bw, XtGrabNone );
	SetSize( NULL, NULL, NULL );
	set_cursor_color( 1 );
	sprintf( (char *)s, "%s [%s]", BOARDTITLE, PATCHLEVEL );
	XSetStandardProperties( XtDisplay(bw), XtWindow(bw),
			(char *)s, BOARDICON, (int)NULL, NULL, 0, NULL );

}



static void Redisplay() {
	GbRedisplayBoard (goban);
}

XtActionsRec actions[] = {
	{ "inputStone", (XtActionProc)InputStone },
	{ "removeStone", (XtActionProc)RemoveStone },
	{ "addStone", (XtActionProc)AlternateStone },
	{ "string", (XtActionProc)HandleStringEvent },
	{ "nline", (XtActionProc)Confirm },
	{ "keyp", (XtActionProc)keyevent },
};

main( argc, argv )
	int argc;
	char **argv;
{
	extern void exit ();
	extern char *optarg;
	extern int optind;
	int c;
	int xtargc = 0;
	Arg args[8];
	int j;
	extern int getopt();
	Pixmap icon_pixmap = None;
	Pixmap mask_pixmap = None;

	get_cmd();

	initboard( boardsize );
	move = stepmove = (-1);
	currentmove = (-1);
	statusflag = FALSE;

#ifdef DEF_X11R3
	{
	Display *dpy;

	XtToolkitInitialize();
	app_context = XtCreateApplicationContext();
	dpy = XtOpenDisplay(app_context, NULL, "XIgc", "XIgc", NULL, 0,
&argc, argv);
	toplevel = XtAppCreateShell("XIgc", NULL, applicationShellWidgetClass,
		dpy, NULL, 0);
	}
#else
	toplevel = XtAppInitialize( &app_context, "XIgc", NULL, 0, &argc, argv,
			NULL, NULL, 0);
#endif

	/* any more command-line args? */
	while( (c=getopt( argc, argv, "v?c:dsalf:12" )) != (-1) ) {
	    switch( c ) {
		case 'a':
			clickflag = 1;
			break;
		case 'd':	/* debug mode */
		    dbg = TRUE;
		    break;
		case 'l':	/* run locally only */
		    localflag = TRUE;
		    break;
		case 's':	/* stippled board */
		    stippleflag = TRUE;
		    break;
		case 'c':	/* command line for comm program */
		    strcpy( (char *)cmdline, (char *)optarg );
		    cmdflag = TRUE;
		    break;
		case '1':	/* use xigcrc COMMAND1 */
		    cmdflag = COMMAND1;
		    break;
		case '2':	/* use xigcrc COMMAND2 */
		    cmdflag = COMMAND2;
		    break;
		case 'f':	/* load a file locally */
		    localflag = TRUE;
		    strcpy( (char *)sgflocalfile, (char *)optarg );
		    sprintf( (char *)fileline, "sgfload %s\n", (char *)optarg );
		    break;
		case '?':
		case 'v':
		default:
		    fprintf( stderr, "%s: Release %s\n", argv[0], PATCHLEVEL );
	    	    fprintf( stderr, USAGE, argv[0] );
		    exit( (-1) );
		    break;
	    }
	}

	if( optind != argc ) {
	    fprintf( stderr, "%s: Release %s\n", argv[0], PATCHLEVEL );
	    fprintf( stderr, USAGE, argv[0] );
	    exit( (-1) );
	}

	if( !cmdflag ) strcpy( (char *)cmdline, CMD );
	else if( cmdflag == COMMAND1 ) strcpy( (char *)cmdline, (char *)cmd1 );
	else if( cmdflag == COMMAND2 ) strcpy( (char *)cmdline, (char *)cmd2 );

	XtSetArg( args[0], XtNiconPixmap, &icon_pixmap );
	XtGetValues( toplevel, args, 1 );
	if( icon_pixmap == None ) {
	    j = 0;

#ifndef HAS_XPM
	    icon_pixmap = XCreateBitmapFromData( XtDisplay(toplevel),
			XtScreen( toplevel )->root, (char *)xigc_icon_bits,
			 xigc_icon_width, xigc_icon_height);
	    XtSetArg( args[j], XtNiconPixmap, icon_pixmap ); ++j;
	    XtSetValues( toplevel, args, j );
#else
	    XpmCreatePixmapFromData( XtDisplay(toplevel),
			    XtScreen( toplevel )->root, (char *)&xigc_icon_xpm,
			    &icon_pixmap, &mask_pixmap, NULL );
	    XtSetArg( args[j], XtNiconPixmap, icon_pixmap ); ++j;
	    XtSetArg( args[j], XtNiconMask, mask_pixmap ); ++j;
	    XtSetValues( toplevel, args, j );
#endif
	}

	bw = XtCreatePopupShell( "board", topLevelShellWidgetClass, toplevel,
			args, j );
	form = XtCreateManagedWidget( "form", formWidgetClass, toplevel, NULL,
			0);

	j = 0;
	XtSetArg( args[j], XtNtype, XawAsciiString ); ++j;
	XtSetArg( args[j], XtNeditType, XawtextEdit ); ++j;
	XtSetArg( args[j], XtNwrap, XawtextWrapWord ); ++j;
	XtSetArg( args[j], XtNautoFill, True ); ++j;
	XtSetArg( args[j], XtNscrollVertical, XawtextScrollWhenNeeded ); ++j;
	echow = XtCreateManagedWidget( "echow", asciiTextWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNlabel, "Game:               " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	gamew = XtCreateManagedWidget( "gamew", labelWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNlabel, "Hcap/Komi:          " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	hcapw = XtCreateManagedWidget( "hcapw", labelWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNlabel, "White:              " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	whitew = XtCreateManagedWidget( "whitew", labelWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNlabel, "Black:              " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	blackw = XtCreateManagedWidget( "blackw", labelWidgetClass, form,
					    args, j );
	
	j = 0;
	XtSetArg( args[j], XtNlabel, "Captured:           " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	captw = XtCreateManagedWidget( "captw", labelWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNlabel, "Captured:           " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	captb = XtCreateManagedWidget( "captb", labelWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNlabel, "Time:               " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	timew = XtCreateManagedWidget( "timew", labelWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNlabel, "Time:               " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	timeb = XtCreateManagedWidget( "timeb", labelWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNlabel, "Last:                                     " ); ++j;
	XtSetArg( args[j], XtNjustify, XtJustifyLeft ); ++j;
	lastw = XtCreateManagedWidget( "lastw", labelWidgetClass, form,
					    args, j );

	j = 0;
	XtSetArg( args[j], XtNgameSize, 19 ); ++j;
	XtSetArg( args[j], XtNcursor, GbCBlackStone); ++j;
	goban = XtCreateManagedWidget( "goban", gobanWidgetClass, bw, args,
			j);

	whow = XtCreateManagedWidget( "who", commandWidgetClass, form,
			NULL, 0);
	gamesw = XtCreateManagedWidget( "games", commandWidgetClass, form,
			NULL, 0);

	j = 0;
	XtSetArg( args[j], XtNsensitive, False); ++j;
	undow = XtCreateManagedWidget( "undo", commandWidgetClass, form,
			args, j );
	passw = XtCreateManagedWidget( "pass", commandWidgetClass, form,
			args, j);

	j = 0;
	refresh = XtCreateManagedWidget( "refresh", commandWidgetClass, form,
			NULL, 0);
	cursorcolor = XtCreateManagedWidget( "cursorcolor", commandWidgetClass,
			form, NULL, 0);
	clearb = XtCreateManagedWidget( "clearb", commandWidgetClass, form,
			NULL, 0);
	stepf = XtCreateManagedWidget( "stepf", commandWidgetClass, form, NULL,
			0);
	stepb = XtCreateManagedWidget( "stepb", commandWidgetClass, form, NULL,
			0);
	stepff = XtCreateManagedWidget( "stepff", commandWidgetClass, form,NULL,
			0);
	stepbf = XtCreateManagedWidget( "stepbf", commandWidgetClass, form,NULL,
			0);
	refreshstep = XtCreateManagedWidget( "refreshstep", commandWidgetClass,
			form, NULL, 0);

	localrefresh = XtCreateManagedWidget( "localrefresh",
			commandWidgetClass, form, NULL, 0);
	quit = XtCreateManagedWidget( "quit", commandWidgetClass, form, NULL,
			0);
	size = XtCreateManagedWidget( "size", commandWidgetClass, form, NULL,	
			0);
	savegame = XtCreateManagedWidget( "savegame", commandWidgetClass,
			form, NULL, 0);
	savecurrent = XtCreateManagedWidget( "savecurrent", commandWidgetClass,
			form, NULL, 0);
	coordinates = XtCreateManagedWidget( "coordinates", commandWidgetClass,
			form, NULL, 0);
	sgfloadw = XtCreateManagedWidget( "sgfload", commandWidgetClass, form,
			NULL, 0);
	sgfdonew = XtCreateManagedWidget( "sgfdone", commandWidgetClass, form,
			NULL, 0);

	XtAppAddActions (app_context, actions, XtNumber(actions));

	XtAddCallback( whow, XtNcallback, (XtCallbackProc)SendMsg,
			(XP)"who\n" );
	XtAddCallback( gamesw, XtNcallback, (XtCallbackProc)SendMsg,
			(XP)"games\n" );
	XtAddCallback( undow, XtNcallback, (XtCallbackProc)SendMsg,
			(XP)"undo\n" );
	XtAddCallback( passw, XtNcallback, (XtCallbackProc)SendMsg,
			(XP)"pass\n" );
	XtAddCallback( refresh, XtNcallback, (XtCallbackProc)RefreshHost,
			NULL );
	XtAddCallback( cursorcolor, XtNcallback, (XtCallbackProc)CursorColor,
			NULL ); 
	XtAddCallback( clearb, XtNcallback, (XtCallbackProc)ClearBoard, NULL );
	XtAddCallback( stepf, XtNcallback, (XtCallbackProc)StepForward, NULL );
	XtAddCallback( stepb, XtNcallback, (XtCallbackProc)StepBack, NULL );
	XtAddCallback( stepff, XtNcallback, (XtCallbackProc)StepForward, (XP)0x01);
	XtAddCallback( stepbf, XtNcallback, (XtCallbackProc)StepBack, (XP)0x01);
	XtAddCallback( refreshstep, XtNcallback, (XtCallbackProc)RefreshStep,
			NULL );
	XtAddCallback( localrefresh, XtNcallback, (XtCallbackProc)LocalRefresh,
			NULL );
	XtAddCallback( size, XtNcallback, (XtCallbackProc)SetSize, NULL );
	XtAddCallback( savegame, XtNcallback, (XtCallbackProc)SaveGame,
			(XP)REALMOVE );
	XtAddCallback( savecurrent, XtNcallback, (XtCallbackProc)SaveGame,
			(XP)STEPMOVE );
	XtAddCallback( coordinates, XtNcallback, (XtCallbackProc)SetCoordinates,
			NULL );
	XtAddCallback( quit, XtNcallback, (XtCallbackProc)Quit, NULL );
	XtAddCallback( sgfloadw, XtNcallback, (XtCallbackProc)SgfLoad, NULL );
	XtAddCallback( sgfdonew, XtNcallback, (XtCallbackProc)SgfDone, NULL );

	XtRealizeWidget (toplevel);

    XSetStandardProperties( XtDisplay(toplevel), XtWindow(toplevel),
		TITLE, ICONTITLE, (int)NULL, argv, argc, NULL );

	SetupBoard();

	dp = XtCreatePopupShell( "Xigc Dialog", topLevelShellWidgetClass,
				toplevel, NULL, 0 );

	dialog = XtCreateManagedWidget( "dialog", dialogWidgetClass,
				dp, NULL, 0 );
	XawDialogAddButton( dialog, "Load", (XtCallbackProc)Confirm, NULL );
	XawDialogAddButton( dialog, "Cancel", (XtCallbackProc)Cancel, NULL );

	if( !localflag ) {
	    setup_filter( &exec_stdin, &exec_stdout, &exec_stderr );
	    input1id = XtAppAddInput( app_context, (int)fileno(exec_stdout),
		(XP)(XtInputReadMask), read_filter,
		(caddr_t)(int)fileno(exec_stdout) );
	    input2id = XtAppAddInput( app_context, (int)fileno(exec_stderr),
		(XP)(XtInputReadMask), read_filter,
		(caddr_t)(int)fileno(exec_stderr) );
	}
	else {
	    exec_stdin = fopen( "/dev/null", "w" );
	    process_kbd( fileline, exec_stdin );
	}

	input0id = XtAppAddInput( app_context, (int)fileno(stdin),
		(XP)(XtInputReadMask), read_kbd,
		(caddr_t)(int)fileno(stdin) );

	XtAppAddTimeOut( XtWidgetToApplicationContext(toplevel), 1000, Ticker,
			 XtWidgetToApplicationContext(toplevel) );
	if( localflag ) {
		fprintf( stderr, "%s> ", prompt_str );
		fflush( stderr );
	}
	XtAppMainLoop (app_context);
}

Xsetup_p( fr, fw, fe )
	FILE *fr, *fw, *fe;
{
	i1id = XtAppAddInput( app_context, (int)fileno(fw),
		(XP)(XtInputReadMask), read_pipe,
		(caddr_t)(int)fileno(fw) );
	i2id = XtAppAddInput( app_context, (int)fileno(fe),
		(XP)(XtInputReadMask), read_pipe,
		(caddr_t)(int)fileno(fe) );
	if( dbg ) {
	    fprintf( stderr, "Xsetup_p: i1id=0x%x, i2id=0x%x\n", i1id, i2id );
	    fflush( stderr );
	}
}

Xclose_p() {
	XtRemoveInput( i1id );
	XtRemoveInput( i2id );
}

void read_pipe( client_data, source, id )
	int client_data;
	int *source;
	char *id;
{
	unsigned char str[LINESIZE+1];
	int ret;

	ret = read( client_data, str, LINESIZE );
	if( ret == (-1) ) perror( "read_pipe" );

	if( dbg ) {
	    fprintf( stderr, "ENTER read_pipe [%d]\n", ret );
	    fflush( stderr );
	}

	if( ret <= 0 ) return;
	else str[ret] = '\0';
	
	fprintf( stderr, "%s", str );
	fflush( stderr );
}
	

/* this one reads input from the comm channel, prints it, and also
 * calls process_input for parsing
 */
void read_filter( client_data, source, id )
	int client_data;
	int *source;
	char *id;
{
	unsigned char str[LINESIZE+1], *s;
	int ret, i;

	ret = read( client_data, str, LINESIZE );
	if( ret <= 0 ) {
/* these lines don't really do anything very useful,
 * and they cause problems on some HP 9000 systems
	    XtRemoveInput( input1id );
	    XtRemoveInput( input2id );
	    fprintf( stderr, "Closing pipes to comm process\n" );
SC*/
	    return;
	}
	else str[ret] = '\0';

	if( dbg ) fprintf( stderr,
		"********* in read_filter() fd=%d **********\n", client_data );

#ifdef CHARS_7_BIT
	/* just check once more to be sure we have only 7-bit chars */
	for( i=0; i<ret; ++i ) str[i] &= 0x7f;
#endif

	process_input( str );

	if( dbg ) fprintf( stderr,
		"######### leave read_filter() [%d chars] #########\n", ret );
}

/* this one just reads lines from the kbd and passes them to comm stdin */
void read_kbd( client_data, source, id )
	int client_data;
	int *source;
	char *id;
{
	unsigned char str[LINESIZE+1];
	int ret;

	watchdog = 0;

	if( dbg ) fprintf( stderr,
		"********* in read_kbd() fd=%d **********\n", client_data );
	ret = read( client_data, str, LINESIZE );
	if( ret <= 0 ) {
	    fprintf( stderr, "\nKBD: no input!  (ignored)\n" );
	    return;
	}
	else str[ret] = '\0';

	/*XXX got to do something about the CTRL-P to make it work right
	 * when cursor is in xterm window
	 */

	process_kbd( str, exec_stdin );
	if( localflag ) {
		fprintf( stderr, "%s> ", prompt_str );
		fflush( stderr );
	}

	if( dbg ) fprintf( stderr,
		"######### leave read_kbd() [%d chars] #########\n", ret );
}

setMark( i, j, t )
	int i, j;
	unsigned char t;
{
	GbSetMark( goban, (Position)i, (Position)j, t, 0 );
}

drawStone( i, h, t, flag )
	int i, h;
	int t;	/* 0=remove, 1=black, 2=white */
	unsigned char flag;
{
	if( dbg ) fprintf( stderr, "Drawing stone at %d,%d  (%d)\n", i, h, t );
	GbSetPoint( goban, i, h, t, flag );
}

update_status( g, r, f, black, white, bcapt, wcapt, btime, wtime, bbyo,
	wbyo, rankw, rankb )
	int g;
	char r, f;
	unsigned char *white, *black;
	int wcapt, bcapt, wtime, btime;
	int bbyo, wbyo;
	char *rankw, *rankb;
{
	Arg args[8];
	int j;
	unsigned char str[80];

	update_gameno( g, r, f );

	if( g < 0 ) sprintf( (char *)str, "Black:" );
	else if( (int)strlen( (char *)rankb ) <= 0 ) 
	    sprintf( (char *)str, "Black: %s", black );
	else sprintf( (char *)str, "Black: %s    %s", black, rankb );
	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( blackw, args, j );

	if( g < 0 ) sprintf( (char *)str, "White:" );
	else if( (int)strlen( (char *)rankw ) <= 0 ) 
	    sprintf( (char *)str, "White: %s", white );
	else sprintf( (char *)str, "White: %s    %s", white, rankw );
	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( whitew, args, j );

	if( g < 0 ) sprintf( (char *)str, "Captured:" );
	else sprintf( (char *)str, "Captured:   %d", wcapt );
	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( captw, args, j );

	if( g < 0 ) sprintf( (char *)str, "Captured:" );
	else sprintf( (char *)str, "Captured:   %d", bcapt );
	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( captb, args, j );

	fix_wtime( wtime, wbyo );
	fix_btime( btime, bbyo );
}

update_move( last )
	unsigned char *last;
{
	int i, j, cnt;
	unsigned char str[30];
	Arg args[8];

	strcpy( (char *)str, "Last:  " );
	j = strlen( str );
	if( last != NULL ) {
		for( i = 0, cnt = 0; i < 30 &&
				last[i] != '\n' &&
				last[i] != '\0' ; ++i ) {
			if( last[i] == ' ' && ++cnt > 2 ) break;
			else str[j++] = last[i];
		}
		str[j] = '\0';
	}

	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( lastw, args, j );
}

update_handicap( hcap, komi )
	int hcap;
	float komi;
{
	Arg args[8];
	int j;
	unsigned char str[80];

	if( hcap < 0 ) sprintf( (char *)str, "Hcap/Komi:" );
	else sprintf( (char *)str, "Hcap/Komi:    %d   /   %2.1f", hcap, komi );
	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( hcapw, args, j );
}

update_gameno( g, r, f )
	int g;
	char r, f;
{
	Arg args[8];
	int j;
	unsigned char str[80];

	if( g < 0 ) sprintf( (char *)str, "Game:" );
	else sprintf( (char *)str, "Game:   %d        [%c %c]", g, r, f );
	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( gamew, args, j );
}

/* get the server config file, with command line to execute, if any
 * return (-1) if no
 *	1) a file $HOME/.xigcrc
 *	2) in this file:
 *		COMMAND    cmd line to execute
 *		COMMAND1   cmd line to execute
 *		COMMAND2   cmd line to execute
 *		QUITCOMMAND cmd to quit from "cmd"
 *		SAVEFORMAT  SGF or TEXT format for "Save game" menu item
 *		TERSEMODE true or false
 *		SAVEDIR full path of directory to save games into
 *		UPDATETIME interval (seconds) between timer ticks
 *		SORTWHO sorts "who" output by player rank
 *		ADDKILL name    Add name to kill list (permanently)
 *		ADDFRIEND name  Add name to kill list (permanently)
 *		TOGK             Activate kill list
 *		TOGF             Activate friends list
 #		TIMERBEEP	time (seconds) to beep before time expires
 #				[when playing only]; no beep if not set
 #		TIMERCOLORTIME  time (seconds) remaining when time display highlights

 */
static int get_cmd() {
	unsigned char *ret;
	FILE *fp;
	unsigned char str[80], s[20];
	int cnt = 0;
	extern char *getenv();
	int len;

	cmd1[0] = '\0';
	cmd2[0] = '\0';
	quitcmd[0] = '\0';
	strcpy( (char *)savedir, "." );

	ret = (unsigned char *)getenv( "HOME" );
	if( ret == NULL ) return( cnt );

	sprintf( (char *)str, "%s/.xigcrc", ret );
	if( dbg ) fprintf( stderr, "config file: >>%s<<\n", str );
	fp = fopen( (char *)str, "r" );
	if( fp  == NULL ) {
	    if( dbg ) fprintf( stderr, "can't open config file >>%s<<\n", str );
	    return( cnt );
	}
	
	while( fgets( (char *)str, 80, fp ) != NULL ) {
	    len = strlen( (char *)str );
	    if( len < 4 || str[0] == '#' ) continue;
	    str[len-1] = '\0';
	    if( dbg ) fprintf( stderr, ">>%s<<\n", str );
	    if( strncmp( (char *)str, "COMMAND ", 8 ) == 0 ) {
			if( cmd1[0] == '\0' ) {
		    	strcpy( (char *)cmd1, (char *)&str[8] );
		    	if( dbg ) fprintf( stderr, "get_cmd: cmd1 >>%s<<\n" , cmd1 );
				cmdflag = COMMAND1;
		    	++cnt;
			}
	    }
	    if( strncmp( (char *)str, "COMMAND1", 8 ) == 0 ) {
			if( cmd1[0] == '\0' ) {
		    	strcpy( (char *)cmd1, (char *)&str[9] );
		    	if( dbg ) fprintf( stderr, "get_cmd: cmd1 >>%s<<\n" , cmd1 );
				cmdflag = COMMAND1;
		    	++cnt;
			}
	    }
	    if( strncmp( (char *)str, "COMMAND2", 8 ) == 0 ) {
			if( cmd2[0] == '\0' ) {
		    	strcpy( (char *)cmd2, (char *)&str[9] );
		    	if( dbg ) fprintf( stderr, "get_cmd: cmd2 >>%s<<\n" , cmd2 );
				if( cmdflag != COMMAND1 ) cmdflag = COMMAND2;
		    	++cnt;
			}
	    }
	    if( strncmp( (char *)str, "QUITCOMMAND", 11 ) == 0 ) {
		if( quitcmd[0] == '\0' ) {
		    strcpy( (char *)quitcmd, (char *)&str[12] );
		    if( dbg ) fprintf( stderr, "get_cmd: quitcmd >>%s<<\n",
					quitcmd );
		    ++cnt;
	        }
	    }
	    if( strncmp( (char *)str, "TERSEMODE", 9 ) == 0 ) {
		    if( strncmp( (char *)&str[10], "TRUE", 4 ) == 0 ||
		        strncmp( (char *)&str[10], "true", 4 ) == 0 )
				tersemode = TRUE;
		    else if( strncmp( (char *)&str[10], "FALSE", 5 ) == 0 ||
		        strncmp( (char *)&str[10], "false", 5 ) == 0 )
				tersemode = FALSE;
		    ++cnt;
	    }
	    if( strncmp( (char *)str, "UPDATETIME", 10 ) == 0 ) {
		    int tmp = 0;
		    if((tmp = atoi((char *)&str[11])) >= 1) update_int = tmp;
		    ++cnt;
	    }
	    if( strncmp( (char *)str, "TIMERCOLORTIME", 14 ) == 0 ) {
		    int tmp = 0;
		    if((tmp = atoi((char *)&str[14])) >= 0) timer_color = tmp;
		    else timer_color = 0;
		    ++cnt;
	    }
	    if( strncmp( (char *)str, "TIMERBEEP", 9 ) == 0 ) {
		    int tmp = 0;
		    if((tmp = atoi((char *)&str[10])) >= 1) timer_beep = tmp;
		    ++cnt;
	    }
	    if( strncmp( (char *)str, "SORTWHO", 7 ) == 0 ) {
		    if( strncmp( (char *)&str[8], "TRUE", 4 ) == 0 ||
		        strncmp( (char *)&str[8], "true", 4 ) == 0 )
				sortwhomode = TRUE;
		    else if( strncmp( (char *)&str[8], "FALSE", 5 ) == 0 ||
		        strncmp( (char *)&str[8], "false", 5 ) == 0 )
				sortwhomode = FALSE;
		    ++cnt;
	    }
	    if( strncmp( (char *)str, "SAVEFORMAT", 10 ) == 0 ) {
		    if( strncmp( (char *)&str[11], "TEXT", 4 ) == 0 ||
		        strncmp( (char *)&str[11], "text", 4 ) == 0 )
				saveformat = TEXT;
		    else if( strncmp( (char *)&str[11], "SGF", 3 ) == 0 ||
		        strncmp( (char *)&str[11], "sgf", 3 ) == 0 )
				saveformat = SGF;
		    ++cnt;
	    }
	    if( strncmp( (char *)str, "SAVEDIR", 7 ) == 0 ) {
		strcpy( (char *)savedir, (char *)&str[8] );
		++cnt;
	    }
	    if( strncmp( (char *)str, "TOGK", 4 ) == 0 ) {
		filterstatus |= 2;
		fprintf( stderr, "LOCAL: Enable kill list\n" );
		++cnt;
	    }
	    if( strncmp( (char *)str, "TOGF", 4 ) == 0 ) {
		filterstatus |= 1;
		fprintf( stderr, "LOCAL: Enable friends list\n" );
		++cnt;
	    }
	    if( strncmp( (char *)str, "ADDKILL", 7 ) == 0 ) {
		sscanf( (char *)str, "%*s %s", s );
		dokill( s );
		++cnt;
	    }
	    if( strncmp( (char *)str, "ADDFRIEND", 9 ) == 0 ) {
		sscanf( (char *)str, "%*s %s", s );
		dofriend( s );
		++cnt;
	    }
	    if (strncmp((char *)str, "FASTSTEPSIZE", 12) == 0) {
		int in_fast_step_size;
		sscanf( (char *)str, "%*s %d", &in_fast_step_size );
		if (in_fast_step_size > 1) fast_step_size = in_fast_step_size;
		++cnt;
	    }
	}

	fclose( fp );
	return( cnt );
}

fix_wtime( wtime, wbyo )
	int wtime;
	int wbyo;
{
	Arg args[8];
	int j;
	unsigned char str[80];
	int atime;
	unsigned char s;

	if( wtime < 0 ) s = '-';
	else s = ' ';
	atime = abs( wtime );

	if( rules == IRULES || rules == PRULES ) {
	    if( wbyo >= 0 ) {
	        sprintf( (char *)str, "Byo Yomi:   %c%d:%02.2d    %d",
		    	s, atime/60, atime%60, wbyo );

			if( atime/60 == 0 && atime%60 < timer_color ) {
				if( !hilitew ) {
					highlight( timew, TRUE );
					hilitew = TRUE;
				}
			}
			else if( hilitew ) {
				highlight( timew, FALSE );
				hilitew = FALSE;
			}
	    }
	    else {
	        sprintf( (char *)str, "Time:   %c%d:%02.2d", s, atime/60, atime%60 );
	    }
	}
	else {
	    if( wbyo >= 0 ) sprintf( (char *)str, "Overtime #%d:   %c%d:%02.2d",
					(wbyo / 2 ), s, atime/60, atime%60 );
	    else sprintf( (char *)str, "Time:   %c%d:%02.2d",
					s, atime/60, atime%60 );
	}
	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( timew, args, j );
}

fix_btime( btime, bbyo )
	int btime, bbyo;
{
	Arg args[8];
	int j;
	unsigned char str[80];
	int atime;
	unsigned char s;

	if( btime < 0 ) s = '-';
	else s = ' ';
	atime = abs( btime );

	if( rules == IRULES || rules == PRULES ) {
	    if( bbyo >= 0 ) {
	        sprintf( (char *)str, "Byo Yomi:   %c%d:%02.2d    %d",
		    	s, atime/60, atime%60, bbyo );

			if( atime/60 == 0 && atime%60 < timer_color ) {
				if( !hiliteb ) {
					highlight( timeb, TRUE );
					hiliteb = TRUE;
				}
			}
			else if( hiliteb ) {
				highlight( timeb, FALSE );
				hiliteb = FALSE;
			}
	    }
	    else {
	        sprintf( (char *)str, "Time:   %c%d:%02.2d", s, atime/60, atime%60 );
	    }
	}
	else {
	    if( bbyo >= 0 ) sprintf( (char *)str, "Overtime #%d:   %c%d:%02.2d",
					(bbyo / 2 ), s, atime/60, atime%60 );
	    else sprintf( (char *)str, "Time:   %c%d:%02.2d",
					s, atime/60, atime%60 );
	}
	j = 0;
	XtSetArg( args[j], XtNlabel, str ); ++j;
	XtSetValues( timeb, args, j );
}


makehandicap( i, j, num, t, type, drawit )
    int i, j, num, drawit;
    int t;
    int type;       /* update REALMOVE, STEPMOVE data? */
{
    if (t == (int)GbWhiteStone) {
      if (type == REALMOVE) { 
        maingame.who = Black;
        addastone(&maingame,j-1+(maingame.width*(i-1)),
	    drawit,TRUE,Handistone,((rules==IRULES)?RULES:GOE_RULES));
      } else {
        stepgame.who = Black;
        if (addastone(&stepgame,j-1+(stepgame.width*(i-1)),
	    drawit,TRUE,Handistone,((rules==IRULES)?RULES:GOE_RULES)) >= 0)
	  stepgame.sgf.gametype[stepgame.stoneNumber] = STEPGAME;
      }

    } else if (t == (int)GbBlackStone) {
      if (type == REALMOVE) {
        maingame.who = White;
        addastone(&maingame,j-1+(maingame.width*(i-1)),
	    drawit,TRUE,Handistone,((rules==IRULES)?RULES:GOE_RULES));
      } else {
        stepgame.who = White;
        if (addastone(&stepgame,j-1+(stepgame.width*(i-1)),
	    drawit,TRUE,Handistone,((rules==IRULES)?RULES:GOE_RULES)) >= 0)
	  stepgame.sgf.gametype[stepgame.stoneNumber] = STEPGAME;
      }
    } else if (t == EMPTY) {
      if (type == REALMOVE) { 
        passplay(&maingame);
      } else {
        passplay(&stepgame);
      }
    }
}

/* test to see if a stone is on position i,j of board REALMOVE or STEPMOVE */
isastone( i, j, type )
	int i, j;
	int type;
{
	if( type == REALMOVE ) {
	    return( anystone( &maingame, j-1+(maingame.width*(i-1)) ) );
	}
	else {
	    return( anystone( &stepgame, j-1+(maingame.width*(i-1)) ) );
	}
}

makemove( i, j, num, t, type, drawit, rules )
    int i, j, num, drawit, rules;
    int t;	    /* color */
    int type;       /* update REALMOVE, STEPMOVE data? */
{
    if (t == (int)GbWhiteStone) {
      if (type == REALMOVE) { 
        maingame.who = Black;
        addastone(&maingame,j-1+(maingame.width*(i-1)),
	    drawit,TRUE,Playstone,rules);
      } else {
        stepgame.who = Black;
        if (addastone(&stepgame,j-1+(stepgame.width*(i-1)),
	    drawit,TRUE,Playstone,rules) >= 0)
	  stepgame.sgf.gametype[stepgame.stoneNumber] = STEPGAME;
      }

    } else if (t == (int)GbBlackStone) {
      if (type == REALMOVE) {
        maingame.who = White;
        addastone(&maingame,j-1+(maingame.width*(i-1)),
	    drawit,TRUE,Playstone,rules);
      } else {
        stepgame.who = White;
        if (addastone(&stepgame,j-1+(stepgame.width*(i-1)),
	    drawit,TRUE,Playstone,rules) >= 0)
	  stepgame.sgf.gametype[stepgame.stoneNumber] = STEPGAME;
      }
    } else if (t == EMPTY) {
      if (type == REALMOVE) { 
        passplay(&maingame);
      } else {
        passplay(&stepgame);
      }
    }

    /* Beep if a real game move comes in but it can't be drawn because the 
     * user is busy with the stepgame.
     */
    if (!drawit && (type == REALMOVE) && (stepgame.stoneNumber < Maxhistory)) 
      beep();
}

/* ARGSUSED */
void HandleStringEvent( w, event, params, nparams )
    Widget w;
    XEvent *event;
    String *params;
    Cardinal *nparams;
{
    if( *nparams != 1 ) return;

    if( (*params)[0] == '0' && (*params)[1] == 'x' && (*params)[2] != '\0' ) {
	char c, *p, hexval[2];
	hexval[0] = hexval[1] = 0;
	for( p = *params+2; (c = *p); p++ ) {
	    hexval[0] *= 16;
	    if( isupper(c) ) c = tolower(c);
	    if( c >= '0' && c <= '9' ) hexval[0] += c - '0';
	    else if( c >= 'a' && c <= 'f' ) hexval[0] += c - 'a' + 10;
	    else break;
	}
	if( c == '\0' ) StringInput( hexval, 1);
    }
    else {
	StringInput( *params, strlen(*params) );
    }
}

void StringInput( string, nbytes )
    char *string;
    int nbytes;
{
	string[nbytes] = '\0';
	fprintf( stderr, "%s", string );
	fflush( stderr );
	process_kbd( string, exec_stdin );
	if( localflag ) {
		fprintf( stderr, "%s> ", prompt_str );
		fflush( stderr );
	}
}
